import { chainToPlatform, circle, } from "@wormhole-foundation/sdk-base";
import { toNative, } from "@wormhole-foundation/sdk-definitions";
import { CONFIG, DEFAULT_TASK_TIMEOUT } from "./config";
import { CircleTransfer } from "./protocols/cctpTransfer";
import { TokenTransfer } from "./protocols/tokenTransfer";
import { getCircleAttestationWithRetry } from "./circle-api";
import { retry } from "./tasks";
import { getTransactionStatusWithRetry, getVaaByTxHashWithRetry, getVaaBytesWithRetry, getVaaWithRetry, } from "./whscan-api";
export class Wormhole {
    _network;
    _platforms;
    _chains;
    config;
    constructor(network, platforms, config) {
        this._network = network;
        this.config = config ?? CONFIG[network];
        this._chains = new Map();
        this._platforms = new Map();
        for (const p of platforms) {
            this._platforms.set(p._platform, new p(network));
        }
    }
    get network() {
        return this._network;
    }
    /**
     * Creates a CircleTransfer object to move Native USDC from one chain to another
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the CircleTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    async circleTransfer(amount, from, to, automatic, payload, nativeGas) {
        if (automatic && payload)
            throw new Error("Payload with automatic delivery is not supported");
        if (!circle.isCircleChain(from.chain) ||
            !circle.isCircleChain(to.chain) ||
            !circle.isCircleSupported(this.network, from.chain) ||
            !circle.isCircleSupported(this.network, to.chain))
            throw new Error(`Network and chain not supported: ${this.network} ${from.chain} `);
        // ensure the amount is > fee + native gas
        if (automatic) {
            const acb = await this.getChain(from.chain).getAutomaticCircleBridge();
            const relayerFee = await acb.getRelayerFee(to.chain);
            const minAmount = relayerFee + (nativeGas ? nativeGas : 0n);
            if (amount < minAmount)
                throw new Error(`Amount must be > ${minAmount} (relayerFee + nativeGas)`);
        }
        return await CircleTransfer.from(this, {
            amount,
            from,
            to,
            automatic,
            payload,
            nativeGas,
        });
    }
    /**
     * Creates a TokenTransfer object to move a token from one chain to another
     * @param token the token to transfer
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the TokenTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    async tokenTransfer(token, amount, from, to, automatic, payload, nativeGas) {
        // TODO: check if `toChain` is gateway supported
        // not enough to check if its a Cosmos chain since Terra/Xpla/Sei are not supported
        // if (chainToPlatform(to.chain) === 'Cosmwasm' ) {
        //   return await GatewayTransfer.from(this, {
        //     token,
        //     amount,
        //     from,
        //     to,
        //     payload,
        //     nativeGas,
        //   });
        // }
        return await TokenTransfer.from(this, {
            token,
            amount,
            from,
            to,
            automatic,
            payload,
            nativeGas,
        });
    }
    /**
     * Gets the contract addresses for a given chain
     * @param chain the chain name or chain id
     * @returns the contract addresses
     */
    getContracts(chain) {
        return this.config.chains[chain]?.contracts;
    }
    /**
     * Returns the platform object, i.e. the class with platform-specific logic and methods
     * @param chain the chain name or platform name
     * @returns the platform context class
     * @throws Errors if platform is not found
     */
    getPlatform(platformName) {
        const platform = this._platforms.get(platformName);
        if (!platform)
            throw new Error(`Not able to retrieve platform ${platform}`);
        return platform;
    }
    /**
     * Returns the chain "context", i.e. the class with chain-specific logic and methods
     * @param chain the chain name
     * @returns the chain context class
     * @throws Errors if context is not found
     */
    getChain(chain) {
        const platform = chainToPlatform(chain);
        if (!this._chains.has(chain))
            this._chains.set(chain, this.getPlatform(platform).getChain(chain));
        return this._chains.get(chain);
    }
    /**
     * Gets the TokenId for a token representation on any chain
     *  These are the Wormhole wrapped token addresses, not necessarily
     *  the cannonical version of that token
     *
     * @param chain The chain name or id to get the wrapped token address
     * @param tokenId The Token ID (chain/address) of the original token
     * @returns The TokenId on the given chain, null if it does not exist
     * @throws Errors if the chain is not supported or the token does not exist
     */
    async getWrappedAsset(chain, token) {
        const ctx = this.getChain(chain);
        const tb = await ctx.getTokenBridge();
        return { chain, address: await tb.getWrappedAsset(token) };
    }
    /**
     *  Taking a  the original TokenId for some wrapped token  chain
     *  These are the Wormhole wrapped token addresses, not necessarily
     *  the cannonical version of that token
     *
     * @param tokenId The Token ID of the token we're looking up the original asset for
     * @returns The Original TokenId corresponding to the token id passed,
     * @throws Errors if the chain is not supported or the token does not exist
     */
    async getOriginalAsset(token) {
        const ctx = this.getChain(token.chain);
        const tb = await ctx.getTokenBridge();
        return await tb.getOriginalAsset(token.address);
    }
    /**
     * Gets the number of decimals for a token on a given chain
     *
     * @param chain The chain name or id of the token/representation
     * @param token The token address
     * @returns The number of decimals
     */
    async getDecimals(chain, token) {
        const ctx = this.getChain(chain);
        return await ctx.getDecimals(token);
    }
    /**
     * Fetches the balance of a given token for a wallet
     *
     * @param walletAddress The wallet address
     * @param tokenId The token ID (its home chain and address on the home chain)
     * @param chain The chain name or id
     * @returns The token balance of the wormhole asset as a BigNumber
     */
    async getBalance(chain, token, walletAddress) {
        const ctx = this.getChain(chain);
        return ctx.getBalance(walletAddress, token);
    }
    /**
     * Gets the associated token account for chains that require it (only Solana currently).
     * @param token the TokenId of the token to get the token account for
     * @param recipient the address of the primary account that may require a separate token account
     * @returns
     */
    async getTokenAccount(recipient, token) {
        return this.getChain(recipient.chain).getTokenAccount(recipient.address, token.address);
    }
    /**
     * Gets the Raw VAA Bytes from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param wormholeMessageId The WormholeMessageId corresponding to the VAA to be fetched
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA bytes if available
     * @throws Errors if the VAA is not available after the retries
     */
    async getVaaBytes(wormholeMessageId, timeout = DEFAULT_TASK_TIMEOUT) {
        return await getVaaBytesWithRetry(this.config.api, wormholeMessageId, timeout);
    }
    /**
     * Gets a VAA from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param wormholeMessageId The WormholeMessageId corresponding to the VAA to be fetched
     * @param decodeAs The VAA type to decode the bytes as
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA if available
     * @throws Errors if the VAA is not available after the retries
     */
    async getVaa(wormholeMessageId, decodeAs, timeout = DEFAULT_TASK_TIMEOUT) {
        return await getVaaWithRetry(this.config.api, wormholeMessageId, decodeAs, timeout);
    }
    /**
     * Gets a VAA from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param txid The Transaction Hash corresponding to the transaction that cause the wormhole core contract to emit a vaa
     * @param decodeAs The VAA type to decode the bytes as
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA if available
     * @throws Errors if the VAA is not available after the retries
     */
    async getVaaByTxHash(txid, decodeAs, timeout = DEFAULT_TASK_TIMEOUT) {
        return await getVaaByTxHashWithRetry(this.config.api, txid, decodeAs, timeout);
    }
    /**
     * Gets the CircleAttestation corresponding to the message hash logged in the transfer transaction.
     * @param msgHash  The keccak256 hash of the message emitted by the circle contract
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The CircleAttestation as a string, if available
     * @throws Errors if the CircleAttestation is not available after the retries
     */
    async getCircleAttestation(msgHash, timeout = DEFAULT_TASK_TIMEOUT) {
        return getCircleAttestationWithRetry(this.config.circleAPI, msgHash, timeout);
    }
    /**
     * Get the status of a transaction, identified by the chain, emitter address, and sequence number
     *
     * @param wormholeMessageId the message id for the Wormhole Message to get transaction status for
     * @returns the TransactionStatus
     */
    async getTransactionStatus(wormholeMessageId, timeout = DEFAULT_TASK_TIMEOUT) {
        return getTransactionStatusWithRetry(this.config.api, wormholeMessageId, timeout);
    }
    /**
     * Parse an address from its canonincal string format to a NativeAddress
     *
     * @param chain The chain the address is for
     * @param address The address in canonical string format
     * @returns The address in the NativeAddress format
     */
    static parseAddress(chain, address) {
        return toNative(chain, address);
    }
    /**
     * Parse an address from its canonincal string format to a NativeAddress
     *
     * @param chain The chain the address is for
     * @param address The native address in canonical string format
     * @returns The ChainAddress
     */
    static chainAddress(chain, address) {
        return { chain, address: toNative(chain, address) };
    }
    /**
     * Parses all relevant information from a transaction given the sending tx hash and sending chain
     *
     * @param chain The sending chain name or context
     * @param tx The sending transaction hash
     * @returns The parsed WormholeMessageId
     */
    static async parseMessageFromTx(chain, txid, timeout) {
        const task = async () => {
            const msgs = await chain.parseTransaction(txid);
            // possible the node we hit does not have this data yet
            // return null to signal retry
            if (msgs.length === 0)
                return null;
            return msgs;
        };
        const parsed = await retry(task, chain.config.blockTime, timeout, "WormholeCore:ParseMessageFromTransaction");
        if (!parsed)
            throw new Error(`No WormholeMessageId found for ${txid}`);
        if (parsed.length != 1)
            throw new Error(`Expected a single VAA, got ${parsed.length}`);
        return parsed;
    }
}
//# sourceMappingURL=wormhole.js.map